# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals, absolute_import

"""
This module implements classes for generating/parsing Lammps data files.
"""

from six.moves import range
from io import open
import re
from collections import OrderedDict

import numpy as np

from monty.json import MSONable, MontyDecoder

from pymatgen.core.structure import Molecule, Structure

__author__ = 'Kiran Mathew'
__email__ = "kmathew@lbl.gov"
__credits__ = 'Brandon Wood'


class LammpsData(MSONable):
    """
    Basic Lammps data: just the atoms section

    Args:
        box_size (list): [[x_min, x_max], [y_min,y_max], [z_min,z_max]]
        atomic_masses (list): [[atom type, mass],...]
        atoms_data (list): [[atom id, mol id, atom type, charge, x, y, z ...], ... ]
    """

    def __init__(self, box_size, atomic_masses, atoms_data):
        self.box_size = box_size
        self.natoms = len(atoms_data)
        self.natom_types = len(atomic_masses)
        self.atomic_masses = list(atomic_masses)
        self.atoms_data = atoms_data

    def __str__(self):
        """
        string representation of LammpsData

        Returns:
            String representation of the data file
        """
        lines = []
        lines.append("Data file generated by pymatgen\n")
        lines.append("{} atoms\n".format(self.natoms))
        lines.append("{} atom types\n".format(self.natom_types))
        lines.append("{} {} xlo xhi\n{} {} ylo yhi\n{} {} zlo zhi".format(
            self.box_size[0][0], self.box_size[0][1],
            self.box_size[1][0], self.box_size[1][1],
            self.box_size[2][0], self.box_size[2][1]))
        self.set_lines_from_list(lines, "Masses", self.atomic_masses)
        self.set_lines_from_list(lines, "Atoms", self.atoms_data)
        return '\n'.join(lines)

    @staticmethod
    def check_box_size(molecule, box_size):
        """
        Check the box size and if necessary translate the molecule so that
        all the sites are contained within the bounding box.

        Args:
            molecule(Molecule)
            box_size (list): [[x_min, x_max], [y_min, y_max], [z_min, z_max]]
        """
        box_lengths_req = [np.max(molecule.cart_coords[:, i])-np.min(molecule.cart_coords[:, i])
                           for i in range(3)]
        box_lengths = [min_max[1] - min_max[0] for min_max in box_size]
        try:
            np.testing.assert_array_less(box_lengths_req, box_lengths)
        except AssertionError:
            box_size = [[0.0, np.ceil(i*1.1)] for i in box_lengths_req]
            print("Minimum required box lengths {} larger than the provided box lengths{}. "
                  "Resetting the box size to {}".format(
                box_lengths_req, box_lengths, box_size))
        com = molecule.center_of_mass
        new_com = [(side[1] + side[0]) / 2 for side in box_size]
        translate_by = np.array(new_com) - np.array(com)
        molecule.translate_sites(range(len(molecule)), translate_by)
        return box_size

    def write_data_file(self, filename):
        """
        write lammps data input file from the string representation
        of the data.

        Args:
            filename (string): data file name
        """
        with open(filename, 'w') as f:
            f.write(self.__str__())

    @staticmethod
    def get_basic_system_info(structure):
        """
        Return basic system info from the given structure.

        Args:
            structure (Structure)

        Returns:
            number of atoms, number of atom types, box size, mapping
            between the atom id and corresponding atomic masses
        """
        natoms = len(structure)
        natom_types = len(structure.symbol_set)
        elements = structure.composition.elements
        elements = sorted(elements, key=lambda el: el.atomic_mass)
        atomic_masses_dict = OrderedDict(
            [(el.symbol, [i + 1, el.data["Atomic mass"]])
             for i, el in enumerate(elements)])
        return natoms, natom_types, atomic_masses_dict

    @staticmethod
    def get_atoms_data(structure, atomic_masses_dict, set_charge=True):
        """
        return the atoms data:
        atom_id, molecule tag, atom_type, charge(if present else 0), x, y, z.
        The molecule_tag is set to 1(i.e the whole structure corresponds to
        just one molecule). This corresponds to lammps command: "atom_style
        charge"

        Args:
            structure (Structure)
            atomic_masses_dict (dict):
                { atom symbol : [atom_id, atomic mass], ... }
            set_charge (bool): whether or not to set the charge field in Atoms

        Returns:
            [[atom_id, molecule tag, atom_type, charge(if present), x, y, z], ... ]
        """
        atoms_data = []
        for i, site in enumerate(structure):
            atom_type = atomic_masses_dict[site.specie.symbol][0]
            if set_charge:
                if hasattr(site, "charge"):
                    atoms_data.append([i + 1, 1, atom_type, site.charge,
                                       site.x, site.y, site.z])
                else:
                    atoms_data.append([i + 1, 1, atom_type, 0.0,
                                       site.x, site.y, site.z])
            else:
                atoms_data.append([i + 1, 1, atom_type,
                                   site.x, site.y, site.z])
        return atoms_data

    @staticmethod
    def set_lines_from_list(lines, block_name, input_list):
        """
        Append the values from the input list that corresponds to the block
        with name 'block_name' to the list of lines.

        Args:
            lines (list)
            block_name (string): name of the data block,
                e.g. 'Atoms', 'Bonds' etc
            input_list (list): list of values
        """
        if input_list:
            lines.append("\n" + block_name + " \n")
            for ad in input_list:
                lines.append(" ".join([str(x) for x in ad]))

    @staticmethod
    def from_structure(input_structure, box_size, set_charge=True):
        """
        Set LammpsData from the given structure. If the input structure is
        a Structure, it is converted to a molecule. TIf the molecule doesnt fit
        in the input box, the box size is updated based on the max and min site
        coordinates of the molecules.

        Args:
            input_structure (Molecule/Structure)
            box_size (list): [[x_min, x_max], [y_min, y_max], [z_min, z_max]]
            set_charge (bool): whether or not to set the charge field in
            Atoms. If true, the charge will be non-zero only if the
            input_structure has the "charge" site property set.

        Returns:
            LammpsData
        """
        if isinstance(input_structure, Structure):
            input_structure = Molecule.from_sites(input_structure.sites)
        box_size = LammpsData.check_box_size(input_structure, box_size)
        natoms, natom_types, atomic_masses_dict = \
            LammpsData.get_basic_system_info(input_structure.copy())
        atoms_data = LammpsData.get_atoms_data(input_structure,
                                               atomic_masses_dict,
                                               set_charge=set_charge)
        return LammpsData(box_size, atomic_masses_dict.values(), atoms_data)

    @staticmethod
    def from_file(data_file, read_charge=True):
        """
        Return LammpsData object from the data file.
        Note: use this to read in data files that conform with
        atom_style = charge or atomic

        Args:
            data_file (string): data file name
            read_charge (bool): if true, read in data files that conform with
                atom_style = charge else atom_style = atomic

        Returns:
            LammpsData
        """
        atomic_masses = []  # atom_type(starts from 1): mass
        box_size = []
        atoms_data = []
        # atom_id, mol_id, atom_type, charge, x, y, z
        if read_charge:
            atoms_pattern = re.compile(
                "^\s*(\d+)\s+(\d+)\s+(\d+)\s+([0-9eE\.+-]+)\s+("
                "[0-9eE\.+-]+)\s+([0-9eE\.+-]+)\s+([0-9eE\.+-]+)\w*")
        # atom_id, mol_id, atom_type, x, y, z
        else:
            atoms_pattern = re.compile(
                "^\s*(\d+)\s+(\d+)\s+(\d+)\s+([0-9eE\.+-]+)\s+("
                "[0-9eE\.+-]+)\s+([0-9eE\.+-]+)\w*")
        # atom_type, mass
        masses_pattern = re.compile("^\s*(\d+)\s+([0-9\.]+)$")
        box_pattern = re.compile(
            "^([0-9eE\.+-]+)\s+([0-9eE\.+-]+)\s+[xyz]lo\s+[xyz]hi")
        with open(data_file) as df:
            for line in df:
                if masses_pattern.search(line):
                    m = masses_pattern.search(line)
                    atomic_masses.append([int(m.group(1)), float(m.group(2))])
                if box_pattern.search(line):
                    m = box_pattern.search(line)
                    box_size.append([float(m.group(1)), float(m.group(2))])
                m = atoms_pattern.search(line)
                if m:
                    # atom id, mol id, atom type
                    line_data = [int(i) for i in m.groups()[:3]]
                    # charge, x, y, z
                    line_data.extend([float(i) for i in m.groups()[3:]])
                    atoms_data.append(line_data)
        return LammpsData(box_size, atomic_masses, atoms_data)

    def as_dict(self):
        d = MSONable.as_dict(self)
        if hasattr(self, "kwargs"):
            d.update(**self.kwargs)
        return d

    @classmethod
    def from_dict(cls, d):
        decoded = {k: MontyDecoder().process_decoded(v) for k, v in d.items()
                   if not k.startswith("@")}
        return cls(**decoded)


class LammpsForceFieldData(LammpsData):
    """
    Sets Lammps data input file from force field parameters. It is recommended
    that the the convenience method from_forcefield_and_topology be used to
    create the object.

    Args:
        box_size (list): [[x_min,x_max], [y_min,y_max], [z_min,z_max]]
        atomic_masses (list): [ [atom type, atomic mass], ... ]
        pair_coeffs (list): pair coefficients,
            [[unique id, sigma, epsilon ], ... ]
        bond_coeffs (list): bond coefficients,
            [[unique id, value1, value2 ], ... ]
        angle_coeffs (list): angle coefficients,
            [[unique id, value1, value2, value3 ], ... ]
        dihedral_coeffs (list): dihedral coefficients,
            [[unique id, value1, value2, value3, value4], ... ]
        improper_coeffs (list): improper dihedral coefficients,
            [[unique id, value1, value2, value3, value4], ... ]
        atoms_data (list): [[atom id, mol id, atom type, charge, x,y,z, ...], ... ]
        bonds_data (list): [[bond id, bond type, value1, value2], ... ]
        angles_data (list): [[angle id, angle type, value1, value2, value3], ... ]
        dihedrals_data (list):
            [[dihedral id, dihedral type, value1, value2, value3, value4], ... ]
        imdihedrals_data (list):
            [[improper dihedral id, improper dihedral type, value1, value2,
            value3, value4], ... ]
    """

    def __init__(self, box_size, atomic_masses, pair_coeffs, bond_coeffs,
                 angle_coeffs, dihedral_coeffs, improper_coeffs, atoms_data,
                 bonds_data, angles_data, dihedrals_data, imdihedrals_data):
        super(LammpsForceFieldData, self).__init__(box_size, atomic_masses,
                                                   atoms_data)
        # number of types
        self.nbond_types = len(bond_coeffs)
        self.nangle_types = len(angle_coeffs)
        self.ndih_types = len(dihedral_coeffs)
        self.nimdih_types = len(improper_coeffs)
        # number of parameters
        self.nbonds = len(bonds_data)
        self.nangles = len(angles_data)
        self.ndih = len(dihedrals_data)
        self.nimdihs = len(imdihedrals_data)
        # coefficients
        self.pair_coeffs = pair_coeffs
        self.bond_coeffs = bond_coeffs
        self.angle_coeffs = angle_coeffs
        self.dihedral_coeffs = dihedral_coeffs
        self.improper_coeffs = improper_coeffs
        # data
        self.bonds_data = bonds_data
        self.angles_data = angles_data
        self.dihedrals_data = dihedrals_data
        self.imdihedrals_data = imdihedrals_data

    def __str__(self):
        """
        returns a string of lammps data input file
        """
        lines = []
        # title
        lines.append("Data file generated by pymatgen\n")

        # count
        lines.append("{} atoms".format(self.natoms))
        lines.append("{} bonds".format(self.nbonds))
        lines.append("{} angles".format(self.nangles))
        if self.ndih > 0:
            lines.append("{} dihedrals".format(self.ndih))
        if self.nimdihs > 0:
            lines.append("{} impropers".format(self.nimdihs))

        # types
        lines.append("\n{} atom types".format(self.natom_types))
        lines.append("{} bond types".format(self.nbond_types))
        lines.append("{} angle types".format(self.nangle_types))
        if self.ndih > 0:
            lines.append("{} dihedral types".format(self.ndih_types))
        if self.nimdihs > 0:
            lines.append("{} improper types".format(self.nimdih_types))

        # box size
        lines.append("\n{} {} xlo xhi\n{} {} ylo yhi\n{} {} zlo zhi".format(
            self.box_size[0][0], self.box_size[0][1],
            self.box_size[1][0], self.box_size[1][1],
            self.box_size[2][0], self.box_size[2][1]))

        # masses
        self.set_lines_from_list(lines, "Masses", self.atomic_masses)

        # coefficients
        self.set_lines_from_list(lines, "Pair Coeffs", self.pair_coeffs)
        self.set_lines_from_list(lines, "Bond Coeffs", self.bond_coeffs)
        self.set_lines_from_list(lines, "Angle Coeffs", self.angle_coeffs)
        if self.ndih > 0:
            self.set_lines_from_list(lines, "Dihedral Coeffs",
                                     self.dihedral_coeffs)
        if self.nimdihs > 0:
            self.set_lines_from_list(lines, "Improper Coeffs",
                                     self.improper_coeffs)

        # data
        self.set_lines_from_list(lines, "Atoms", self.atoms_data)
        self.set_lines_from_list(lines, "Bonds", self.bonds_data)
        self.set_lines_from_list(lines, "Angles", self.angles_data)
        if self.ndih > 0:
            self.set_lines_from_list(lines, "Dihedrals", self.dihedrals_data)
        if self.nimdihs > 0:
            self.set_lines_from_list(lines, "Impropers", self.imdihedrals_data)
        return '\n'.join(lines)

    @staticmethod
    def get_param_coeff(forcefield, param_name):
        """
        get the parameter coefficients and mapping from the force field.

        Args:
            forcefield (ForceField): ForceField object
            param_name (string): name of the parameter for which
            the coefficients are to be set.

        Returns:
            [[parameter id, value1, value2, ... ], ... ] and
            {parameter key: parameter id, ...}
        """
        if hasattr(forcefield, param_name):
            param = getattr(forcefield, param_name)
            param_coeffs = []
            param_map = {}
            if param:
                for i, item in enumerate(param.items()):
                    param_coeffs.append([i + 1] + list(item[1]))
                    param_map[item[0]] = i+1
            return param_coeffs, param_map
        else:
            raise AttributeError

    @staticmethod
    def get_atoms_data(mols, mols_number, molecule, atomic_masses_dict, topologies):
        """
        Return the atoms data.

        Args:
            mols (list): list of Molecule objects.
            mols_number (list): number of each type of molecule in mols list.
            molecule (Molecule): the molecule assembled from the molecules
                in the mols list.
            topologies (list): list of Topology objects, one for each molecule
                type in mols list

        Returns:
            atoms_data: [[atom id, mol type, atom type, charge, x, y, z], ... ]
            molid_to_atomid: [ [global atom id 1, id 2, ..], ...], the
                index will be the global mol id
        """
        atom_to_mol = {}
        molid_to_atomid = []
        atoms_data = []
        nmols = len(mols)
        # set up map atom_id --> [mol_type, local atom id in the mol] in mols
        # set up map gobal molecule id --> [[atom_id,...],...]
        shift_ = 0
        for mol_type in range(nmols):
            natoms = len(mols[mol_type])
            for num_mol_id in range(mols_number[mol_type]):
                tmp = []
                for mol_atom_id in range(natoms):
                    atom_id = num_mol_id * natoms + mol_atom_id + shift_
                    atom_to_mol[atom_id] = [mol_type, mol_atom_id]
                    tmp.append(atom_id)
                molid_to_atomid.append(tmp)
            shift_ += len(mols[mol_type]) * mols_number[mol_type]
        # set atoms data from the molecule assembly consisting of
        # molecules from mols list with their count from mol_number list.
        # atom id, mol id, atom type, charge from topology, x, y, z
        for i, site in enumerate(molecule):
            atom_type = atomic_masses_dict[site.specie.symbol][0]
            # atom_type = molecule.symbol_set.index(site.species_string) + 1
            atom_id = i + 1
            mol_type = atom_to_mol[i][0] + 1
            mol_atom_id = atom_to_mol[i][1] + 1
            charge = 0.0
            if hasattr(topologies[0], "charges"):
                if topologies[mol_type - 1].charges:
                    charge = topologies[mol_type - 1].charges[mol_atom_id - 1]
            atoms_data.append([atom_id, mol_type, atom_type, charge,
                               site.x, site.y, site.z])
        return atoms_data, molid_to_atomid

    @staticmethod
    def get_param_data(param_name, param_map, mols, mols_number, topologies,
                       molid_to_atomid):
        """
        set the data for the parameter named param_name from the topology.

        Args:
            param_name (string): parameter name, example: "bonds"
            param_map (dict):
                { mol_type: {parameter_key : unique parameter id, ... }, ... }
                example: {0: {("c1","c2"): 1}} ==> c1-c2 bond in mol_type=0
                    has the global id of 1
            mols (list): list of molecules.
            mols_number (list): number of each type of molecule in mols list.
            topologies (list): list of Topology objects, one for each molecule
                type in mols list
            molid_to_atomid (list): [ [gloabal atom id 1, id 2, ..], ...],
                the index is the global mol id

        Returns:
            [ [parameter id, parameter type, global atom id1, global atom id2, ...], ... ]
        """
        param_data = []
        if hasattr(topologies[0], param_name) and getattr(topologies[0], param_name):
            nmols = len(mols)
            mol_id = 0
            # set the map param_to_mol:
            # {global param_id :[global mol id, mol_type, local param id in the param], ... }
            param_to_mol = {}
            shift_ = 0
            for mol_type in range(nmols):
                param_obj = getattr(topologies[mol_type], param_name)
                nparams = len(param_obj)
                for num_mol_id in range(mols_number[mol_type]):
                    mol_id += 1
                    for mol_param_id in range(nparams):
                        param_id = num_mol_id * nparams + mol_param_id + shift_
                        param_to_mol[param_id] = [mol_id - 1, mol_type, mol_param_id]
                shift_ += nparams * mols_number[mol_type]
            # set the parameter data using the topology info
            # example: loop over all bonds in the system
            skip = 0
            for param_id, pinfo in param_to_mol.items():
                mol_id = pinfo[0]  # global molecule id
                mol_type = pinfo[1]  # type of molecule
                mol_param_id = pinfo[2]  # local parameter id in that molecule
                # example: get the bonds list for mol_type molecule
                param_obj = getattr(topologies[mol_type], param_name)
                # connectivity info(local atom ids and type) for the parameter with the local id
                # 'mol_param_id'. example: single bond = [i, j, bond_type]
                param = param_obj[mol_param_id]
                param_atomids = []
                # loop over local atom ids that constitute the parameter
                # for the molecule type, mol_type
                # example: single bond = [i,j,bond_label]
                for atomid in param[:-1]:
                    # local atom id to global atom id
                    global_atom_id = molid_to_atomid[mol_id][atomid]
                    param_atomids.append(global_atom_id + 1)
                param_type = param[-1]
                param_type_reversed = tuple(reversed(param_type))
                # example: get the unique number id for the bond_type
                if param_type in param_map:
                    key = param_type
                elif param_type_reversed in param_map:
                    key = param_type_reversed
                else:
                    key = None
                if key:
                    param_type_id = param_map[key]
                    param_data.append([param_id + 1 - skip, param_type_id] + param_atomids)
                else:
                    skip += 1
                    print("{} or {} Not available".format(param_type, param_type_reversed))
        return param_data

    @staticmethod
    def from_forcefield_and_topology(mols, mols_number, box_size, molecule,
                                     forcefield, topologies):
        """
        Return LammpsForceFieldData object from force field and topology info
        for the 'molecule' assembled from the constituent molecules specified
        in the 'mols' list with their count specified in the 'mols_number' list.

        Args:
            mols (list): List of Molecule objects
            mols_number (list): List of number of molecules of each
                molecule type in mols
            box_size (list): [[x_min,x_max], [y_min,y_max], [z_min,z_max]]
            molecule (Molecule): The molecule that is assembled from mols
                and mols_number
            forcefield (ForceFiled): Force filed information
            topologies (list): List of Topology objects, one for each
                molecule type in mols.

        Returns:
            LammpsForceFieldData
        """
        # set the coefficients and map from the force field
        bond_coeffs, bond_map = LammpsForceFieldData.get_param_coeff(
            forcefield, "bonds")
        angle_coeffs, angle_map = LammpsForceFieldData.get_param_coeff(
            forcefield, "angles")
        pair_coeffs, _ = LammpsForceFieldData.get_param_coeff(
            forcefield, "pairs")
        dihedral_coeffs, dihedral_map = LammpsForceFieldData.get_param_coeff(
            forcefield, "dihedrals")
        improper_coeffs, imdihedral_map = LammpsForceFieldData.get_param_coeff(
            forcefield, "imdihedrals")
        # atoms data, topology used for setting charge if present
        box_size = LammpsForceFieldData.check_box_size(molecule, box_size)
        natoms, natom_types, atomic_masses_dict = \
            LammpsData.get_basic_system_info(molecule.copy())
        atoms_data, molid_to_atomid = LammpsForceFieldData.get_atoms_data(
            mols, mols_number, molecule, atomic_masses_dict, topologies)
        # set the other data from the molecular topologies
        bonds_data = LammpsForceFieldData.get_param_data(
            "bonds", bond_map, mols, mols_number, topologies, molid_to_atomid)
        angles_data = LammpsForceFieldData.get_param_data(
            "angles", angle_map, mols, mols_number, topologies, molid_to_atomid)
        dihedrals_data = LammpsForceFieldData.get_param_data(
            "dihedrals", dihedral_map, mols, mols_number, topologies, molid_to_atomid)
        imdihedrals_data = LammpsForceFieldData.get_param_data(
            "imdihedrals", imdihedral_map, mols, mols_number, topologies, molid_to_atomid)
        return LammpsForceFieldData(box_size, atomic_masses_dict.values(),
                                    pair_coeffs, bond_coeffs,
                                    angle_coeffs, dihedral_coeffs,
                                    improper_coeffs, atoms_data,
                                    bonds_data, angles_data, dihedrals_data,
                                    imdihedrals_data)

    @staticmethod
    def from_file(data_file):
        """
        Return LammpsForceFieldData object from the data file. It is assumed
        that the forcefield paramter sections for pairs, bonds, angles,
        dihedrals and improper dihedrals are named as follows(not case sensitive):
        "Pair Coeffs", "Bond Coeffs", "Angle Coeffs", "Dihedral Coeffs" and
        "Improper Coeffs". For "Pair Coeffs", values for factorial(n_atom_types)
        pairs must be specified.

        Args:
            data_file (string): the data file name

        Returns:
            LammpsForceFieldData
        """
        atomic_masses = []  # atom_type(starts from 1): mass
        box_size = []
        pair_coeffs = []
        bond_coeffs = []
        angle_coeffs = []
        dihedral_coeffs = []
        improper_coeffs = []
        atoms_data = []
        bonds_data = []
        angles_data = []
        dihedral_data = []
        imdihedral_data = []
        types_pattern = re.compile("^\s*(\d+)\s+([a-zA-Z]+)\s+types$")
        # atom_id, mol_id, atom_type, charge, x, y, z
        atoms_pattern = re.compile(
            "^\s*(\d+)\s+(\d+)\s+(\d+)\s+([0-9eE\.+-]+)\s+("
            "[0-9eE\.+-]+)\s+([0-9eE\.+-]+)\s+([0-9eE\.+-]+)\w*")
        masses_pattern = re.compile("^\s*(\d+)\s+([0-9\.]+)$")
        box_pattern = re.compile(
            "^\s*([0-9eE\.+-]+)\s+([0-9eE\.+-]+)\s+[xyz]lo\s+[xyz]hi")
        # id, value1, value2
        general_coeff_pattern = re.compile("^\s*(\d+)\s+([0-9\.]+)\s+([0-9\.]+)$")
        # id, type, atom_id1, atom_id2
        bond_data_pattern = re.compile("^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$")
        # id, type, atom_id1, atom_id2, atom_id3
        angle_data_pattern = re.compile(
            "^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$")
        # id, type, atom_id1, atom_id2, atom_id3, atom_id4
        dihedral_data_pattern = re.compile(
            "^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$")
        read_pair_coeffs = False
        read_bond_coeffs = False
        read_angle_coeffs = False
        read_dihedral_coeffs = False
        read_improper_coeffs = False
        with open(data_file) as df:
            for line in df:
                if types_pattern.search(line):
                    m = types_pattern.search(line)
                    if m.group(2) == "atom":
                        natom_types = int(m.group(1))
                        npair_types = natom_types # i != j skipped
                    if m.group(2) == "bond":
                        nbond_types = int(m.group(1))
                    if m.group(2) == "angle":
                        nangle_types = int(m.group(1))
                    if m.group(2) == "dihedral":
                        ndihedral_types = int(m.group(1))
                    if m.group(2) == "improper":
                        nimproper_types = int(m.group(1))
                if masses_pattern.search(line):
                    m = masses_pattern.search(line)
                    atomic_masses.append([int(m.group(1)), float(m.group(2))])
                if box_pattern.search(line):
                    m = box_pattern.search(line)
                    box_size.append([float(m.group(1)), float(m.group(2))])
                if "Pair Coeffs".lower() in line.lower():
                    read_pair_coeffs = True
                    continue
                if read_pair_coeffs:
                    tokens = line.split()
                    if tokens:
                        pair_coeffs.append([int(tokens[0])] + [float(i) for i in tokens[1:]])
                        read_pair_coeffs = False if len(pair_coeffs) >= npair_types else True
                if "Bond Coeffs".lower() in line.lower():
                    read_bond_coeffs = True
                    continue
                if read_bond_coeffs:
                    m = general_coeff_pattern.search(line)
                    if m:
                        bond_coeffs.append(
                            [int(m.group(1)), float(m.group(2)), float(m.group(3))])
                        read_bond_coeffs = False if len(bond_coeffs) >= nbond_types else True
                if "Angle Coeffs".lower() in line.lower():
                    read_angle_coeffs = True
                    continue
                if read_angle_coeffs:
                    m = general_coeff_pattern.search(line)
                    if m:
                        angle_coeffs.append(
                            [int(m.group(1)), float(m.group(2)), float(m.group(3))])
                        read_angle_coeffs = False if len(angle_coeffs) >= nangle_types else True
                if "Dihedral Coeffs".lower() in line.lower():
                    read_dihedral_coeffs = True
                    continue
                if read_dihedral_coeffs:
                    tokens = line.split()
                    if tokens:
                        dihedral_coeffs.append(
                            [int(tokens[0])] + [float(i) for i in tokens[1:]])
                        read_dihedral_coeffs = False if len(dihedral_coeffs) >= ndihedral_types else True
                if "Improper Coeffs".lower() in line.lower():
                    read_improper_coeffs = True
                    continue
                if read_improper_coeffs:
                    tokens = line.split()
                    if tokens:
                        improper_coeffs.append(
                            [int(tokens[0])] + [float(i) for i in tokens[1:]])
                        read_improper_coeffs = False if len(improper_coeffs) >= nimproper_types else True
                if atoms_pattern.search(line):
                    m = atoms_pattern.search(line)
                    # atom id, mol id, atom type
                    line_data = [int(i) for i in m.groups()[:3]]
                    # charge, x, y, z, vx, vy, vz ...
                    line_data.extend([float(i) for i in m.groups()[3:]])
                    atoms_data.append(line_data)
                if bond_data_pattern.search(line) and atoms_data:
                    m = bond_data_pattern.search(line)
                    bonds_data.append([int(i) for i in m.groups()])
                if angle_data_pattern.search(line) and atoms_data:
                    m = angle_data_pattern.search(line)
                    angles_data.append([int(i) for i in m.groups()])
                if dihedral_data_pattern.search(line) and atoms_data:
                    m = dihedral_data_pattern.search(line)
                    dihedral_data.append([int(i) for i in m.groups()])
        return LammpsForceFieldData(box_size, atomic_masses, pair_coeffs,
                                    bond_coeffs, angle_coeffs,
                                    dihedral_coeffs, improper_coeffs,
                                    atoms_data, bonds_data, angles_data,
                                    dihedral_data, imdihedral_data)



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals, absolute_import

"""
This module implements classes for processing Lammps output files:

    1. log file: contains the thermodynamic data with the format set by the
        'thermo_style' command

    2. trajectory file(dump file): the file generated by the 'dump' command
"""

import re
from io import open

import numpy as np

from pymatgen.core.periodic_table import _pt_data
from pymatgen.core.structure import Molecule
from pymatgen.core.lattice import Lattice
from pymatgen.analysis.diffusion_analyzer import DiffusionAnalyzer
from pymatgen.io.lammps.data import LammpsData, LammpsForceFieldData

__author__ = "Kiran Mathew"
__email__ = "kmathew@lbl.gov"
__credits__ = "Navnidhi Rajput, Michael Humbert"


class LammpsRun(object):
    """
    Parse the lammps data file, trajectory file and the log file to extract
    useful info about the system.

    Args:
        data_file (str): path to the data file
        trajectory_file (str): path to the trajectory file
        log_file (str): path to the log file
    """

    def __init__(self, data_file, trajectory_file, log_file="log.lammps",
                 is_forcefield=False):
        self.data_file = data_file
        self.trajectory_file = trajectory_file
        self.log_file = log_file
        self.lammps_log = LammpsLog(log_file)
        if is_forcefield:
            self.lammps_data = LammpsForceFieldData.from_file(data_file)
        else:
            self.lammps_data = LammpsData.from_file(data_file)
        self._set_mol_masses_and_charges()
        self._parse_trajectory()

    def _parse_trajectory(self):
        """
        parse the trajectory file.
        """
        traj_timesteps = []
        trajectory = []
        timestep_label = "ITEM: TIMESTEP"
        # "ITEM: ATOMS id type ...
        traj_label_pattern = re.compile(
            "^\s*ITEM:\s+ATOMS\s+id\s+type\s+([A-Za-z\s]*)")
        # default: id type x y z vx vy vz mol"
        # updated below based on the field names in the ITEM: ATOMS line
        # Note: the first 2 fields must be the id and the atom type. There can be
        # arbitrary number of fields after that and they all will be treated as floats.
        traj_pattern = re.compile(
            "\s*(\d+)\s+(\d+)\s+([0-9eE\.+-]+)\s+([0-9eE\.+-]+)\s+"
            "([0-9eE\.+-]+)\s+"
            "([0-9eE\.+-]+)\s+"
            "([0-9eE\.+-]+)\s+([0-9eE\.+-]+)\s+(\d+)\s*")
        parse_timestep = False
        with open(self.trajectory_file) as tf:
            for line in tf:
                if timestep_label in line:
                    parse_timestep = True
                    continue
                if parse_timestep:
                    traj_timesteps.append(float(line))
                    parse_timestep = False
                if traj_label_pattern.search(line):
                    fields = traj_label_pattern.search(line).group(1)
                    fields = fields.split()
                    # example:- id type x y z vx vy vz mol ...
                    traj_pattern_string = "\s*(\d+)\s+(\d+)" + "".join(
                        ["\s+([0-9eE\.+-]+)" for _ in range(len(fields))])
                    traj_pattern = re.compile(traj_pattern_string)
                if traj_pattern.search(line):
                    # first 2 fields must be id and type, the rest of them
                    # will be casted as floats
                    m = traj_pattern.search(line)
                    line_data = []
                    line_data.append(int(m.group(1)) - 1)
                    line_data.append(int(m.group(2)))
                    line_data.extend(
                        [float(x) for i, x in enumerate(m.groups()) if i + 1 > 2])
                    trajectory.append(tuple(line_data))
        traj_dtype = np.dtype([(str('Atoms_id'), np.int64),
                               (str('atom_type'), np.int64)] +
                              [(str(fld), np.float64) for fld in fields])
        self.trajectory = np.array(trajectory, dtype=traj_dtype)
        self.timesteps = np.array(traj_timesteps, dtype=np.float64)
        for step in range(self.timesteps.size):
            begin = step * self.natoms
            end = (step + 1) * self.natoms
            self.trajectory[begin:end] = np.sort(self.trajectory[begin:end],
                                                 order=str("Atoms_id"))

    def _set_mol_masses_and_charges(self):
        """
        set the charge, mass and the atomic makeup for each molecule
        """
        mol_config = []  # [ [atom id1, atom id2, ...], ... ]
        mol_masses = []  # [ [atom mass1, atom mass2, ...], ... ]
        # mol_charges = []
        unique_atomic_masses = np.array(self.lammps_data.atomic_masses)[:, 1]
        atoms_data = np.array(self.lammps_data.atoms_data)
        mol_ids = atoms_data[:, 1].astype(np.int64)
        atom_ids = atoms_data[:, 0].astype(np.int64)
        unique_mol_ids = np.unique(mol_ids)
        atomic_types = atoms_data[:, 2].astype(np.int64)
        atomic_masses = unique_atomic_masses[atomic_types - 1]
        # atomic_charges = atoms_data[:, 3]
        self.nmols = unique_mol_ids.size
        for umid in range(self.nmols):
            mol_config.append(atom_ids[np.where(mol_ids == umid + 1)] - 1)
            mol_masses.append(atomic_masses[np.where(mol_ids == umid + 1)])
            # mol_charges.append(np.sum(atomic_charges[np.where(mol_ids == umid+1)]))
        self.mol_config = np.array(mol_config)
        self.mol_masses = np.array(mol_masses)

    def _weighted_average(self, mol_id, mol_vector):
        """
        Calculate the weighted average of the array comprising of
        atomic vectors corresponding to the molecule with id mol_id.

        Args:
            mol_id (int): molecule id
            mol_vector (numpy array): array of shape,
                natoms_in_molecule with id mol_id x 3

        Returns:
            1D numpy array(3 x 1) of weighted averages in x, y, z directions
        """
        mol_masses = self.mol_masses[mol_id]
        return np.array([np.dot(mol_vector[:, dim], mol_masses) / np.sum(mol_masses)
                         for dim in range(3)])

    def _get_mol_vector(self, step, mol_id, param=["x", "y", "z"]):
        """
        Returns numpy array corresponding to atomic vectors of parameter
        "param" for the given time step and molecule id

        Args:
            step (int): time step
            mol_id (int): molecule id
            param (list): the atomic parameter for which the weighted
                average is to be computed

        Returns:
            2D numpy array(natoms_in_molecule x 3) of atomic vectors
        """
        begin = step * self.natoms
        end = (step + 1) * self.natoms
        mol_vector_structured = \
            self.trajectory[begin:end][self.mol_config[mol_id]][param]
        new_shape = mol_vector_structured.shape + (-1,)
        mol_vector = mol_vector_structured.view(np.float64).reshape(new_shape)
        return mol_vector.copy()

    # TODO: remove this and use only get_displacements(an order of magnitude faster)
    def get_structures_from_trajectory(self):
        """
        Convert the coordinates in each time step to a structure(boxed molecule).
        Used to construct DiffusionAnalyzer object.

        Returns:
            list of Structure objects
        """
        structures = []
        mass_to_symbol = dict(
            (round(y["Atomic mass"], 1), x) for x, y in _pt_data.items())
        unique_atomic_masses = np.array(self.lammps_data.atomic_masses)[:, 1]
        for step in range(self.timesteps.size):
            begin = step * self.natoms
            end = (step + 1) * self.natoms
            mol_vector_structured = \
                self.trajectory[begin:end][:][["x", "y", "z"]]
            new_shape = mol_vector_structured.shape + (-1,)
            mol_vector = mol_vector_structured.view(np.float64).reshape(
                new_shape)
            coords = mol_vector.copy()
            species = [mass_to_symbol[round(unique_atomic_masses[atype - 1], 1)]
                       for atype in self.trajectory[begin:end][:]["atom_type"]]
            mol = Molecule(species, coords)
            try:
                boxed_mol = mol.get_boxed_structure(*self.box_lengths)
            except ValueError as error:
                print("Error: '{}' at timestep {} in the trajectory".format(
                    error,
                    int(self.timesteps[step])))
            structures.append(boxed_mol)
        return structures

    def get_displacements(self):
        """
        Return the initial structure and displacements for each time step.
        Used to interface witht the DiffusionAnalyzer.

        Returns:
            Structure object, numpy array of displacements
        """
        lattice = Lattice([[self.box_lengths[0], 0, 0],
                           [0, self.box_lengths[1], 0],
                           [0, 0, self.box_lengths[2]]])
        mass_to_symbol = dict(
            (round(y["Atomic mass"], 1), x) for x, y in _pt_data.items())
        unique_atomic_masses = np.array(self.lammps_data.atomic_masses)[:, 1]
        frac_coords = []
        for step in range(self.timesteps.size):
            begin = step * self.natoms
            end = (step + 1) * self.natoms
            mol_vector_structured = \
                self.trajectory[begin:end][:][["x", "y", "z"]]
            new_shape = mol_vector_structured.shape + (-1,)
            mol_vector = mol_vector_structured.view(np.float64).reshape(
                new_shape)
            coords = mol_vector.copy()
            if step == 0:
                species = [
                    mass_to_symbol[round(unique_atomic_masses[atype - 1], 1)]
                    for atype in self.trajectory[begin:end][:]["atom_type"]]
                mol = Molecule(species, coords)
                structure = mol.get_boxed_structure(*self.box_lengths)
            step_frac_coords = [lattice.get_fractional_coords(crd)
                                for crd in coords]
            frac_coords.append(np.array(step_frac_coords)[:, None])
        frac_coords = np.concatenate(frac_coords, axis=1)
        dp = frac_coords[:, 1:] - frac_coords[:, :-1]
        dp = dp - np.round(dp)
        f_disp = np.cumsum(dp, axis=1)
        disp = lattice.get_cartesian_coords(f_disp)
        return structure, disp

    def get_diffusion_analyzer(self, specie, temperature, time_step, step_skip,
                               smoothed=None, min_obs=30, avg_nsteps=1000):
        """
        Args:
            specie (Element/Specie): Specie to calculate diffusivity for as a
                String. E.g., "Li".
            temperature (float): Temperature of the diffusion run in Kelvin.
            time_step (int): Time step between measurements.
            step_skip (int): Sampling frequency of the displacements (
                time_step is multiplied by this number to get the real time
                between measurements)

            For the other parameters please see the
            pymatgen.analysis.diffusion_analyzer.DiffusionAnalyzer documentation.

        Returns:
            DiffusionAnalyzer
        """
        # structures = self.get_structures_from_trajectory()
        structure, disp = self.get_displacements()
        return DiffusionAnalyzer(structure, disp, specie, temperature,
                                 time_step, step_skip=step_skip, smoothed=smoothed,
                                 min_obs=min_obs, avg_nsteps=avg_nsteps)

    @property
    def natoms(self):
        return self.lammps_data.natoms

    @property
    def box_lengths(self):
        return [l[1] - l[0] for l in self.lammps_data.box_size]

    @property
    def traj_timesteps(self):
        """
        trajectory time steps in time units.
        e.g. for units = real, time units = fmsec
        """
        return self.timesteps * self.lammps_log.timestep

    @property
    def mol_trajectory(self):
        """
        Compute the weighted average trajectory of each molecule at each
        timestep

        Returns:
            2D numpy array ((n_timesteps*mols_number) x 3)
        """
        traj = []
        for step in range(self.timesteps.size):
            tmp_mol = []
            for mol_id in range(self.nmols):
                mol_coords = self._get_mol_vector(step, mol_id,
                                                  param=["x", "y", "z"])
                # take care of periodic boundary conditions
                pbc_wrap(mol_coords, self.box_lengths)
                tmp_mol.append(self._weighted_average(mol_id, mol_coords))
            traj.append(tmp_mol)
        return np.array(traj)

    @property
    def mol_velocity(self):
        """
         Compute the weighted average velcoity of each molecule at each
         timestep.

         Returns:
            2D numpy array ((n_timesteps*mols_number) x 3)
        """
        velocity = []
        for step in range(self.timesteps.size):
            tmp_mol = []
            for mol_id in range(self.nmols):
                mol_velocities = self._get_mol_vector(step, mol_id,
                                                      param=["vx", "vy", "vz"])
                tmp_mol.append(self._weighted_average(mol_id, mol_velocities))
            velocity.append(tmp_mol)
        return np.array(velocity)


class LammpsLog(object):
    """
    Parser for LAMMPS log file.
    """

    def __init__(self, log_file="log.lammps"):
        """
        Args:
            log_file (string): path to the loag file
        """
        self.log_file = log_file
        self._parse_log()

    def _parse_log(self):
        """
        Parse the log file for the thermodynamic data.
        Sets the thermodynamic data as a structured numpy array with field names
        taken from the the thermo_style command.
        """
        thermo_data = []
        thermo_pattern = None
        with open(self.log_file, 'r') as logfile:
            for line in logfile:
                # timestep, the unit depedns on the 'units' command
                time = re.search('timestep\s+([0-9]+)', line)
                if time and not thermo_data:
                    self.timestep = float(time.group(1))
                # total number md steps
                steps = re.search('run\s+([0-9]+)', line)
                if steps and not thermo_data:
                    self.nmdsteps = int(steps.group(1))
                # logging interval
                thermo = re.search('thermo\s+([0-9]+)', line)
                if thermo and not thermo_data:
                    self.interval = float(thermo.group(1))
                # thermodynamic data, set by the thermo_style command
                format = re.search('thermo_style.+', line)
                if format and not thermo_data:
                    fields = format.group().split()[2:]
                    thermo_pattern_string = "\s*([0-9eE\.+-]+)" + "".join(
                        ["\s+([0-9eE\.+-]+)" for _ in range(len(fields) - 1)])
                    thermo_pattern = re.compile(thermo_pattern_string)
                if thermo_pattern:
                    if thermo_pattern.search(line):
                        m = thermo_pattern.search(line)
                        thermo_data.append(
                            tuple([float(x) for i, x in enumerate(m.groups())]))
        thermo_data_dtype = np.dtype([(str(fld), np.float64) for fld in fields])
        self.thermo_data = np.array(thermo_data, dtype=thermo_data_dtype)


def pbc_wrap(array, box_lengths):
    """
    wrap the array for molecule coordinates around the periodic boundary.

    Args:
        array (numpy.ndarray): molecule coordinates, [[x1,y1,z1],[x2,y2,z2],..]
        box_lengths (list): [x_length, y_length, z_length]
    """
    ref = array[0, 0]
    for i in range(3):
        array[:, i] = np.where((array[:, i] - ref) >= box_lengths[i] / 2,
                               array[:, i] - box_lengths[i], array[:, i])
        array[:, i] = np.where((array[:, i] - ref) < -box_lengths[i] / 2,
                               array[:, i] + box_lengths[i], array[:, i])



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals, absolute_import

"""
This module defines utility classes and functions.
"""

import six
from io import open
import os
import tempfile
from subprocess import Popen, PIPE

import numpy as np

try:
    import pybel as pb
except ImportError:
    pb = None
from pymatgen import Molecule
from pymatgen.core.operations import SymmOp
from pymatgen.util.coord_utils import get_angle
from pymatgen.io.babel import BabelMolAdaptor

from monty.os.path import which
from monty.dev import requires
from monty.tempfile import ScratchDir

__author__ = 'Kiran Mathew, Brandon Wood, Michael Humbert'
__email__ = 'kmathew@lbl.gov'


class Polymer(object):
    """
    Generate polymer chain via Random walk. At each position there are
    a total of 5 possible moves(excluding the previous direction).
    """
    def __init__(self, start_monomer, s_head, s_tail,
                 monomer, head, tail,
                 end_monomer, e_head, e_tail,
                 n_units, link_distance=1.0, linear_chain=False):
        """
        Args:
            start_monomer (Molecule): Starting molecule
            s_head (int): starting atom index of the start_monomer molecule
            s_tail (int): tail atom index of the start_monomer
            monomer (Molecule): The monomer
            head (int): index of the atom in the monomer that forms the head
            tail (int): tail atom index. monomers will be connected from
                tail to head
            end_monomer (Molecule): Terminal molecule
            e_head (int): starting atom index of the end_monomer molecule
            e_tail (int): tail atom index of the end_monomer
            n_units (int): number of monomer units excluding the start and
                terminal molecules
            link_distance (float): distance between consecutive monomers
            linear_chain (bool): linear or random walk polymer chain
        """
        self.start = s_head
        self.end = s_tail
        self.monomer = monomer
        self.n_units = n_units
        self.link_distance = link_distance
        self.linear_chain = linear_chain
        # translate monomers so that head atom is at the origin
        start_monomer.translate_sites(range(len(start_monomer)),
                                      - monomer.cart_coords[s_head])
        monomer.translate_sites(range(len(monomer)),
                                - monomer.cart_coords[head])
        end_monomer.translate_sites(range(len(end_monomer)),
                                    - monomer.cart_coords[e_head])
        self.mon_vector = monomer.cart_coords[tail] - monomer.cart_coords[head]
        self.moves = {1: [1, 0, 0],
                      2: [0, 1, 0],
                      3: [0, 0, 1],
                      4: [-1, 0, 0],
                      5: [0, -1, 0],
                      6: [0, 0, -1]}
        self.prev_move = 1
        # places the start monomer at the beginning of the chain
        self.molecule = start_monomer.copy()
        self.length = 1
        # create the chain
        self._create(self.monomer, self.mon_vector)
        # terminate the chain with the end_monomer
        self.n_units += 1
        end_mon_vector = end_monomer.cart_coords[e_tail] - \
                         end_monomer.cart_coords[e_head]
        self._create(end_monomer, end_mon_vector)

    def _create(self, monomer, mon_vector):
        """
        create the polymer from the monomer

        Args:
            monomer (Molecule)
            mon_vector (numpy.array): molecule vector that starts from the
                start atom index to the end atom index
        """
        while self.length != (self.n_units-1):
            if self.linear_chain:
                move_direction = np.array(mon_vector) / np.linalg.norm(mon_vector)
            else:
                move_direction = self._next_move_direction()
            self._add_monomer(monomer.copy(), mon_vector, move_direction)

    def _next_move_direction(self):
        """
        pick a move at random from the list of moves
        """
        nmoves = len(self.moves)
        move = np.random.randint(1, nmoves+1)
        while self.prev_move == (move + 3) % nmoves:
            move = np.random.randint(1, nmoves+1)
        self.prev_move = move
        return np.array(self.moves[move])

    def _align_monomer(self, monomer, mon_vector, move_direction):
        """
        rotate the monomer so that it is aligned along the move direction

        Args:
            monomer (Molecule)
            mon_vector (numpy.array): molecule vector that starts from the
                start atom index to the end atom index
            move_direction (numpy.array): the direction of the polymer chain
                extension
        """
        axis = np.cross(mon_vector, move_direction)
        origin = monomer[self.start].coords
        angle = get_angle(mon_vector, move_direction)
        op = SymmOp.from_origin_axis_angle(origin, axis, angle)
        monomer.apply_operation(op)

    def _add_monomer(self, monomer, mon_vector, move_direction):
        """
        extend the polymer molecule by adding a monomer along mon_vector direction

        Args:
            monomer (Molecule): monomer molecule
            mon_vector (numpy.array): monomer vector that points from head to tail.
            move_direction (numpy.array): direction along which the monomer
                will be positioned
        """
        translate_by = self.molecule.cart_coords[self.end] + \
                       self.link_distance * move_direction
        monomer.translate_sites(range(len(monomer)), translate_by)
        if not self.linear_chain:
            self._align_monomer(monomer, mon_vector, move_direction)
        # add monomer if there are no crossings
        does_cross = False
        for i, site in enumerate(monomer):
            try:
                self.molecule.append(site.specie, site.coords,
                                     properties=site.properties)
            except:
                does_cross = True
                polymer_length = len(self.molecule)
                self.molecule.remove_sites(
                    range(polymer_length - i, polymer_length))
                break
        if not does_cross:
            self.length += 1
            self.end += len(self.monomer)


class PackmolRunner(object):
    """
    Wrapper for the Packmol software that can be used to pack various types of
    molecules into a one single unit.
    """

    @requires(which('packmol'),
              "PackmolRunner requires the executable 'packmol' to be in "
              "the path. Please download packmol from "
              "https://github.com/leandromartinez98/packmol "
              "and follow the instructions in the README to compile. "
              "Don't forget to add the packmol binary to your path")
    def __init__(self, mols, param_list, input_file="pack.inp",
                 tolerance=2.0, filetype="xyz",
                 control_params={"maxit": 20, "nloop": 600},
                 auto_box=True, output_file="packed.xyz"):
        """
        Args:
              mols:
                   list of Molecules to pack
              input_file:
                        name of the packmol input file
              tolerance:
                        min distance between the atoms
              filetype:
                       input/output structure file type
              control_params:
                           packmol control parameters dictionary. Basically
                           all parameters other than structure/atoms
              param_list:
                    list of parameters containing dicts for each molecule
              auto_box:
                    put the molecule assembly in a box
              output_file:
                    output file name. The extension will be adjusted
                    according to the filetype
        """
        self.mols = mols
        self.param_list = param_list
        self.input_file = input_file
        self.boxit = auto_box
        self.control_params = control_params
        if not self.control_params.get("tolerance"):
            self.control_params["tolerance"] = tolerance
        if not self.control_params.get("filetype"):
            self.control_params["filetype"] = filetype
        if not self.control_params.get("output"):
            self.control_params["output"] = "{}.{}".format(
                output_file.split(".")[0], self.control_params["filetype"])
        if self.boxit:
            self._set_box()

    def _format_param_val(self, param_val):
        """
        Internal method to format values in the packmol parameter dictionaries

        Args:
              param_val:
                   Some object to turn into String

        Returns:
               string representation of the object
        """
        if isinstance(param_val, list):
            return ' '.join(str(x) for x in param_val)
        else:
            return str(param_val)

    def _set_box(self):
        """
        Set the box size for the molecular assembly
        """
        net_volume = 0.0
        for idx, mol in enumerate(self.mols):
            length = max([np.max(mol.cart_coords[:, i])-np.min(mol.cart_coords[:, i])
                           for i in range(3)]) + 2.0
            net_volume += (length**3.0) * float(self.param_list[idx]['number'])
        length = net_volume**(1.0/3.0)
        for idx, mol in enumerate(self.mols):
            self.param_list[idx]['inside box'] = '0.0 0.0 0.0 {} {} {}'.format(
                length, length, length)

    def _write_input(self, input_dir="."):
        """
        Write the packmol input file to the input directory.

        Args:
            input_dir (string): path to the input directory
        """
        with open(os.path.join(input_dir, self.input_file), 'wt', encoding="utf-8") as inp:
            for k, v in six.iteritems(self.control_params):
                inp.write('{} {}\n'.format(k, self._format_param_val(v)))
            # write the structures of the constituent molecules to file and set
            # the molecule id and the corresponding filename in the packmol
            # input file.
            for idx, mol in enumerate(self.mols):
                a = BabelMolAdaptor(mol)
                pm = pb.Molecule(a.openbabel_mol)
                filename = os.path.join(
                    input_dir, '{}.{}'.format(
                        idx, self.control_params["filetype"])).encode("ascii")
                pm.write(self.control_params["filetype"], filename=filename,
                         overwrite=True)
                inp.write("\n")
                inp.write(
                    "structure {}.{}\n".format(
                        os.path.join(input_dir, str(idx)),
                        self.control_params["filetype"]))
                for k, v in six.iteritems(self.param_list[idx]):
                    inp.write('  {} {}\n'.format(k, self._format_param_val(v)))
                inp.write('end structure\n')

    def run(self, copy_to_current_on_exit=False):
        """
        Write the input file to the scratch directory, run packmol and return
        the packed molecule.

        Args:
            copy_to_current_on_exit (bool): Whether or not to copy the packmol
                input/output files from the scratch directory to the current
                directory.

        Returns:
                Molecule object
        """
        scratch = tempfile.gettempdir()
        with ScratchDir(scratch, copy_to_current_on_exit=copy_to_current_on_exit) as scratch_dir:
            self._write_input(input_dir=scratch_dir)
            packmol_bin = ['packmol']
            packmol_input = open(os.path.join(scratch_dir, self.input_file), 'r')
            p = Popen(packmol_bin, stdin=packmol_input, stdout=PIPE, stderr=PIPE)
            p.wait()
            (stdout, stderr) = p.communicate()
            output_file = os.path.join(scratch_dir, self.control_params["output"])
            if os.path.isfile(output_file):
                packed_mol = BabelMolAdaptor.from_file(output_file)
                print("packed molecule written to {}".format(
                    self.control_params["output"]))
                return packed_mol.pymatgen_mol
            else:
                print("Packmol execution failed")
                print(stdout, stderr)
                return None


class LammpsRunner(object):
    def __init__(self, dict_input, input_filename="lammps.in", bin="lammps"):
        """
        LAMMPS wrapper

        Args:
            dict_input (DictLammpsInput): lammps input object
            input_filename (string): input file name
            bin (string): command to run, excluding the input file name
        """
        self.lammps_bin = bin.split()
        if not which(self.lammps_bin[-1]):
            raise RuntimeError(
                "LammpsRunner requires the executable {} to be in the path. "
                "Please download and install LAMMPS from " \
                "http://lammps.sandia.gov. "
                "Don't forget to add the binary to your path".format(self.lammps_bin[-1]))
        self.dict_input = dict_input
        self.input_filename = input_filename

    def run(self):
        """
        Write the input/data files and run LAMMPS.
        """
        self.dict_input.write_input(self.input_filename)
        print("Input file: {}".format(self.input_filename))
        lammps_cmd = self.lammps_bin + ['-in', self.input_filename]
        print("Running: {}".format(" ".join(lammps_cmd)))
        p = Popen(lammps_cmd, stdout=PIPE, stderr=PIPE)
        p.wait()
        (stdout, stderr) = p.communicate()
        print("Done")
        print(stdout, stderr)


if __name__ == '__main__':
    ethanol_coords = [[0.00720, -0.56870, 0.00000],
                      [-1.28540, 0.24990, 0.00000],
                      [1.13040, 0.31470, 0.00000],
                      [0.03920, -1.19720, 0.89000],
                      [0.03920, -1.19720, -0.89000],
                      [-1.31750, 0.87840, 0.89000],
                      [-1.31750, 0.87840, -0.89000],
                      [-2.14220, -0.42390, -0.00000],
                      [1.98570, -0.13650, -0.00000]]
    ethanol = Molecule(["C", "C", "O", "H", "H", "H", "H", "H", "H"],
                       ethanol_coords)
    water_coords = [[9.626, 6.787, 12.673],
                    [9.626, 8.420, 12.673],
                    [10.203, 7.604, 12.673]]
    water = Molecule(["H", "H", "O"], water_coords)
    pmr = PackmolRunner([ethanol, water],
                        [{"number": 1, "fixed": [0, 0, 0, 0, 0, 0],
                          "centerofmass": ""},
                         {"number": 15, "inside sphere": [0, 0, 0, 5]}],
                        input_file="packmol_input.inp", tolerance=2.0,
                        filetype="xyz",
                        control_params={"nloop": 1000},
                        auto_box=False, output_file="cocktail.xyz")
    s = pmr.run()



#!/usr/bin/env python


__author__ = 'waroquiers'

import unittest2 as unittest
import os
import json
import shutil
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import LocalGeometryFinder
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import StructureEnvironments
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import LightStructureEnvironments
from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import SimplestChemenvStrategy
from pymatgen.core.structure import Structure

json_files_dir = os.path.join(os.path.dirname(__file__), "..", "..", "..", "..", "..",
                              'test_files', "chemenv", "json_test_files")


class ReadWriteChemenvTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.lgf = LocalGeometryFinder()
        cls.lgf.setup_parameters(centering_type='standard')
        os.makedirs('tmp_dir')

    def test_read_structure_environments(self):
        f = open("{}/{}".format(json_files_dir, 'test_T--4_FePO4_icsd_4266.json'), 'r')
        dd = json.load(f)
        f.close()

        atom_indices = dd['atom_indices']

        struct = Structure.from_dict(dd['structure'])
        self.lgf.setup_structure(struct)
        se = self.lgf.compute_structure_environments_detailed_voronoi(only_indices=atom_indices,
                                                                      maximum_distance_factor=2.25)

        f = open('tmp_dir/se.json', 'w')
        json.dump(se.as_dict(), f)
        f.close()

        f = open('tmp_dir/se.json', 'r')
        dd = json.load(f)
        f.close()

        se2 = StructureEnvironments.from_dict(dd)

        self.assertEqual(se, se2)

        _strategy = SimplestChemenvStrategy()
        light_se = LightStructureEnvironments(_strategy, se)

        f = open('tmp_dir/light_se.json', 'w')
        json.dump(light_se.as_dict(), f)
        f.close()

        f = open('tmp_dir/light_se.json', 'r')
        dd = json.load(f)
        f.close()

        light_se2 = LightStructureEnvironments.from_dict(dd)

        self.assertEqual(light_se._strategy, light_se2._strategy)
        self.assertEqual(light_se._structure, light_se2._structure)
        self.assertEqual(light_se._bva_valences, light_se2._bva_valences)
        self.assertEqual(light_se._coordination_environments, light_se2._coordination_environments)
        self.assertEqual(light_se._neighbors, light_se2._neighbors)
        self.assertEqual(light_se, light_se2)

    @classmethod
    def tearDownClass(cls):
        #Remove the directory in which the temporary files have been created
        shutil.rmtree('tmp_dir')


if __name__ == "__main__":
    unittest.main()


# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module provides input and output mechanisms
for the xr file format, which is a modified CSSR
file format and, for example, used in GULP.
In particular, the module makes it easy
to remove shell positions from relaxations
that employed core-shell models.
"""


__author__ = "Nils Edvin Richard Zimmermann"
__copyright__ = "Copyright 2016, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Nils Edvin Richard Zimmermann"
__email__ = "nils.e.r.zimmermann@gmail.com"
__date__ = "June 23, 2016"

import re

from six.moves import map
import numpy as np

from monty.io import zopen
from math import fabs
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure

class Xr(object):
    """
    Basic object for working with xr files.

    Args:
        structure (Structure/IStructure): Structure object to create the
                Xr object.
    """

    def __init__(self, structure):
        if not structure.is_ordered:
            raise ValueError("Xr file can only be constructed from ordered "
                             "structure")
        self.structure = structure

    def __str__(self):
        output = ["pymatgen   {:.4f} {:.4f} {:.4f}"
                  .format(*self.structure.lattice.abc),
                  "{:.3f} {:.3f} {:.3f}"
                  .format(*self.structure.lattice.angles),
                  "{} 0".format(len(self.structure)),
                  "0 {}".format(self.structure.formula)]
        # There are actually 10 more fields per site
        # in a typical xr file from GULP, for example.
        for i, site in enumerate(self.structure.sites):
            output.append("{} {} {:.4f} {:.4f} {:.4f}"
                          .format(i + 1, site.specie, site.x, site.y, site.z))
        mat = self.structure.lattice.matrix
        for i in range(2):
            for j in range(3):
                output.append("{:.4f} {:.4f} {:.4f}".format(
                        mat[j][0], mat[j][1], mat[j][2]))
        return "\n".join(output)

    def write_file(self, filename):
        """
        Write out an xr file.

        Args:
            filename (str): name of the file to write to.
        """
        with zopen(filename, 'wt') as f:
            f.write(str(self) + "\n")

    @staticmethod
    def from_string(string, use_cores=True, thresh=1.e-4):
        """
        Creates an Xr object from a string representation.

        Args:
            string (str): string representation of an Xr object.
            use_cores (bool): use core positions and discard shell
                    positions if set to True (default).  Otherwise,
                    use shell positions and discard core positions.
            thresh (float): relative threshold for consistency check
                    between cell parameters (lengths and angles) from
                    header information and cell vectors, respectively.

        Returns:
            xr (Xr): Xr object corresponding to the input
                    string representation.
        """
        lines = string.split("\n")
        toks = lines[0].split()
        lengths = [float(toks[i]) for i in range(1, len(toks))]
        toks = lines[1].split()
        angles = [float(i) for i in toks[0:3]]
        toks = lines[2].split()
        nsites = int(toks[0])
        mat = np.zeros((3,3), dtype=float)
        for i in range(3):
            toks = lines[4+nsites+i].split()
            toks2 = lines[4+nsites+i+3].split()
            for j, item in enumerate(toks):
                if item != toks2[j]:
                    raise RuntimeError("expected both matrices"
                            " to be the same in xr file")
            mat[i] = np.array([float(w) for w in toks])
        lat = Lattice(mat)
        if fabs(lat.a-lengths[0])/fabs(lat.a) > thresh or \
                fabs(lat.b-lengths[1])/fabs(lat.b) > thresh or \
                fabs(lat.c-lengths[2])/fabs(lat.c) > thresh or \
                fabs(lat.alpha-angles[0])/fabs(lat.alpha) > thresh or \
                fabs(lat.beta-angles[1])/fabs(lat.beta) > thresh or \
                fabs(lat.gamma-angles[2])/fabs(lat.gamma) > thresh:
            raise RuntimeError("cell parameters in header ("+str(lengths)+\
                    ", "+str(angles)+") are not consistent with Cartesian"+\
                    " lattice vectors ("+str(lat.abc)+", "+\
                    str(lat.angles)+")")
        # Ignore line w/ index 3.
        sp = []
        coords = []
        for j in range(nsites):
            m = re.match("\d+\s+(\w+)\s+([0-9\-\.]+)\s+([0-9\-\.]+)\s+" +
                         "([0-9\-\.]+)", lines[4+j].strip())
            if m:
                tmp_sp = m.group(1)
                if use_cores and tmp_sp[len(tmp_sp)-2:] == "_s":
                    continue
                if not use_cores and tmp_sp[len(tmp_sp)-2:] == "_c":
                    continue
                if tmp_sp[len(tmp_sp)-2] == "_":
                    sp.append(tmp_sp[0:len(tmp_sp)-2])
                else:
                    sp.append(tmp_sp)
                coords.append([float(m.group(i)) for i in range(2, 5)])
        return Xr(Structure(lat, sp, coords, coords_are_cartesian=True))

    @staticmethod
    def from_file(filename, use_cores=True, thresh=1.e-4):
        """
        Reads an xr-formatted file to create an Xr object.

        Args:
            filename (str): name of file to read from.
            use_cores (bool): use core positions and discard shell
                    positions if set to True (default).  Otherwise,
                    use shell positions and discard core positions.
            thresh (float): relative threshold for consistency check
                    between cell parameters (lengths and angles) from
                    header information and cell vectors, respectively.

        Returns:
            xr (Xr): Xr object corresponding to the input
                    file.
        """
        with zopen(filename, "rt") as f:
            return Xr.from_string(
                    f.read(), use_cores=use_cores,
                    thresh=thresh)




# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.
"""Tools to compute equations of states with different models."""
from __future__ import unicode_literals, division, print_function

import collections
import numpy as np
import pymatgen.core.units as units

from monty.functools import return_none_if_raise
from pymatgen.core.units import FloatWithUnit
from pymatgen.util.plotting_utils import add_fig_kwargs, get_ax_fig_plt

import logging
logger = logging.getLogger(__file__)

__all__ = [
    "EOS",
]


def quadratic(V, a, b, c):
    """Quadratic fit"""
    return a*V**2 + b*V + c


def murnaghan(V, E0, B0, B1, V0):
    """From PRB 28,5480 (1983)"""

    E = E0 + B0*V/B1*(((V0/V)**B1)/(B1-1)+1) - V0*B0/(B1-1)
    return E


def birch(V, E0, B0, B1, V0):
    """
    From Intermetallic compounds: Principles and Practice, Vol. I: Principles
    Chapter 9 pages 195-210 by M. Mehl. B. Klein, D. Papaconstantopoulos paper downloaded from Web

    case where n=0
    """

    E = (E0
         + 9.0/8.0*B0*V0*((V0/V)**(2.0/3.0) - 1.0)**2
         + 9.0/16.0*B0*V0*(B1-4.)*((V0/V)**(2.0/3.0) - 1.0)**3)
    return E


def birch_murnaghan(V, E0, B0, B1, V0):
    """BirchMurnaghan equation from PRB 70, 224107"""

    eta = (V/V0)**(1./3.)
    E = E0 + 9.*B0*V0/16.*(eta**2-1)**2*(6 + B1*(eta**2-1.) - 4.*eta**2)
    return E


def pourier_tarantola(V, E0, B0, B1, V0):
    """Pourier-Tarantola equation from PRB 70, 224107"""

    eta = (V/V0)**(1./3.)
    squiggle = -3.*np.log(eta)

    E = E0 + B0*V0*squiggle**2/6.*(3. + squiggle*(B1 - 2))
    return E


def vinet(V, E0, B0, B1, V0):
    'Vinet equation from PRB 70, 224107'

    eta = (V/V0)**(1./3.)

    E = (E0 + 2.*B0*V0/(B1-1.)**2
         * (2. - (5. +3.*B1*(eta-1.)-3.*eta)*np.exp(-3.*(B1-1.)*(eta-1.)/2.)))
    return E


def deltafactor_polyfit(volumes, energies):
    """
    This is the routine used to compute V0, B0, B1 in the deltafactor code.

    Taken from deltafactor/eosfit.py
    """
    fitdata = np.polyfit(volumes**(-2./3.), energies, 3, full=True)
    ssr = fitdata[1]
    sst = np.sum((energies - np.average(energies))**2.)
    residuals0 = ssr/sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    v0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            v0 = x**(-3./2.)
            break
    else:
        raise EOSError("No minimum could be found")

    derivV2 = 4./9. * x**5. * deriv2(x)
    derivV3 = (-20./9. * x**(13./2.) * deriv2(x) - 8./27. * x**(15./2.) * deriv3(x))
    b0 = derivV2 / x**(3./2.)
    b1 = -1 - x**(-3./2.) * derivV3 / derivV2

    #print('deltafactor polyfit:')
    #print('e0, b0, b1, v0')
    #print(fitdata[0], b0, b1, v0)

    n = collections.namedtuple("DeltaFitResults", "v0 b0 b1 poly1d")
    return n(v0, b0, b1, fitdata[0])



class EOSError(Exception):
    """Exceptions raised by EOS."""


class EOS(object):
    """
    Fit equation of state for bulk systems.

    The following equation is used::

       murnaghan
           PRB 28, 5480 (1983)

       birch
           Intermetallic compounds: Principles and Practice, Vol I: Principles. pages 195-210

       birch_murnaghan
           PRB 70, 224107

       pourier_tarantola
           PRB 70, 224107

       vinet
           PRB 70, 224107

    Use::

       eos = EOS(eos_name='murnaghan')
       fit = eos.fit(volumes, energies)
       print(fit)
       fit.plot()

    """
    Error = EOSError

    #: Models available.
    MODELS = {
        "quadratic": quadratic,
        "murnaghan": murnaghan,
        "birch": birch,
        "birch_murnaghan": birch_murnaghan,
        "pourier_tarantola": pourier_tarantola,
        "vinet": vinet,
        "deltafactor": deltafactor_polyfit,
    }

    def __init__(self, eos_name='murnaghan'):
        self._eos_name = eos_name
        self._func = self.MODELS[eos_name]

    @staticmethod
    def Quadratic():
        return EOS(eos_name="quadratic")

    @staticmethod
    def Murnaghan():
        return EOS(eos_name='murnaghan')

    @staticmethod
    def Birch():
        return EOS(eos_name='birch')

    @staticmethod
    def Birch_Murnaghan():
        return EOS(eos_name='birch_murnaghan')

    @staticmethod
    def Pourier_Tarantola():
        return EOS(eos_name='pourier_tarantola')

    @staticmethod
    def Vinet():
        return EOS(eos_name='vinet')

    @staticmethod
    def DeltaFactor():
        return EOS(eos_name='deltafactor')

    def fit(self, volumes, energies, vol_unit="ang^3", ene_unit="eV"):
        """
        Fit energies [eV] as function of volumes [Angstrom**3].

        Returns `EosFit` instance that gives access to the optimal volume,
        the minumum energy, and the bulk modulus.
        Notice that the units for the bulk modulus is eV/Angstrom^3.
        """
        # Convert volumes to Ang**3 and energies to eV (if needed).
        volumes = units.ArrayWithUnit(volumes, vol_unit).to("ang^3")
        energies = units.EnergyArray(energies, ene_unit).to("eV")

        return EOS_Fit(volumes, energies, self._func, self._eos_name)



class EOS_Fit(object):
    """Performs the fit of E(V) and provides method to access the results of the fit."""

    def __init__(self, volumes, energies, func, eos_name):
        """
        args:
            energies: list of energies in eV
            volumes: list of volumes in Angstrom^3
            func: callable function
        """
        self.volumes = np.array(volumes)
        self.energies = np.array(energies)
        assert len(self.volumes) == len(self.energies)

        self.func = func
        self.eos_name = eos_name
        self.exceptions = []
        self.ierr = 0

        if eos_name == "deltafactor":
            try:
                results = deltafactor_polyfit(self.volumes, self.energies)

                self.e0 = None
                self.v0 = results.v0
                self.b0 = results.b0
                self.b1 = results.b1
                self.p0 = results.poly1d
                self.eos_params = results.poly1d

            except EOSError as exc:
                self.ierr = 1
                logger.critical(str(exc))
                self.exceptions.append(exc)
                raise

        elif eos_name == "quadratic":
            # Quadratic fit
            a, b, c = np.polyfit(self.volumes, self.energies, 2)

            self.v0 = v0 = -b/(2*a)
            self.e0 = a*v0**2 + b*v0 + c
            self.b0 = 2*a*v0
            self.b1 = np.inf
            self.p0 = [a, b, c]
            self.eos_params = [a, b, c]

            vmin, vmax = self.volumes.min(), self.volumes.max()

            if not vmin < v0 and v0 < vmax:
                exc = EOSError('The minimum volume of a fitted parabola is not in the input volumes\n.')
                logger.critical(str(exc))
                self.exceptions.append(exc)

        else:
            # Objective function that will be minimized
            def objective(pars, x, y):
                return y - self.func(x, *pars)

            # Quadratic fit to get an initial guess for the parameters
            a, b, c = np.polyfit(self.volumes, self.energies, 2)

            v0 = -b/(2*a)
            e0 = a*v0**2 + b*v0 + c
            b0 = 2*a*v0
            b1 = 4  # b1 is usually a small number like 4

            vmin, vmax = self.volumes.min(), self.volumes.max()

            if not vmin < v0 and v0 < vmax:
                exc = EOSError('The minimum volume of a fitted parabola is not in the input volumes\n.')
                logger.critical(str(exc))
                self.exceptions.append(exc)

            # Initial guesses for the parameters
            self.p0 = [e0, b0, b1, v0]

            from scipy.optimize import leastsq
            self.eos_params, self.ierr = leastsq(objective, self.p0, args=(self.volumes, self.energies))

            if self.ierr not in [1, 2, 3, 4]:
                exc = EOSError("Optimal parameters not found")
                logger.critical(str(exc))
                self.exceptions.append(exc)
                raise exc

            self.e0 = self.eos_params[0]
            self.b0 = self.eos_params[1]
            self.b1 = self.eos_params[2]
            self.v0 = self.eos_params[3]

            print('EOS_fit:', func)
            print('e0, b0, b1, v0')
            print(self.eos_params)

    def __str__(self):
        lines = []
        app = lines.append
        app("Equation of State: %s" % self.name)
        app("Minimum volume = %1.2f Ang^3" % self.v0)
        app("Bulk modulus = %1.2f eV/Ang^3 = %1.2f GPa, b1 = %1.2f" % (self.b0, self.b0_GPa, self.b1))

        return "\n".join(lines)

    @property
    def name(self):
        return self.func.__name__

    @property
    def b0_GPa(self):
        return FloatWithUnit(self.b0, "eV ang^-3").to("GPa")

    @property
    @return_none_if_raise(AttributeError)
    def results(self):
        """Dictionary with the results. None if results are not available"""
        return dict(v0=self.v0, e0=self.e0, b0=self.b0, b1=self.b1)

    @add_fig_kwargs
    def plot(self, ax=None, **kwargs):
        """
        Uses Matplotlib to plot the energy curve.

        Args:
            ax: :class:`Axes` object. If ax is None, a new figure is produced.

        ================  ==============================================================
        kwargs            Meaning
        ================  ==============================================================
        style             
        color
        text
        label
        ================  ==============================================================

        Returns:
            Matplotlib figure.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        vmin, vmax = self.volumes.min(), self.volumes.max()
        emin, emax = self.energies.min(), self.energies.max()

        vmin, vmax = (vmin - 0.01 * abs(vmin), vmax + 0.01 * abs(vmax))
        emin, emax = (emin - 0.01 * abs(emin), emax + 0.01 * abs(emax))

        color = kwargs.pop("color", "r")
        label = kwargs.pop("label", None)

        # Plot input data.
        ax.plot(self.volumes, self.energies, linestyle="None", marker="o", color=color) #, label="Input Data")

        # Plot EOS.
        vfit = np.linspace(vmin, vmax, 100)
        if label is None:
            label = self.name + ' fit'

        if self.eos_name == "deltafactor":
            xx = vfit**(-2./3.)
            ax.plot(vfit, np.polyval(self.eos_params, xx), linestyle="dashed", color=color, label=label)
        else:
            ax.plot(vfit, self.func(vfit, *self.eos_params), linestyle="dashed", color=color, label=label)

        # Set xticks and labels.
        ax.grid(True)
        ax.set_xlabel("Volume $\AA^3$")
        ax.set_ylabel("Energy (eV)")

        ax.legend(loc="best", shadow=True)

        # Add text with fit parameters.
        if kwargs.pop("text", True):
            text = []; app = text.append
            app("Min Volume = %1.2f $\AA^3$" % self.v0)
            app("Bulk modulus = %1.2f eV/$\AA^3$ = %1.2f GPa" % (self.b0, self.b0_GPa))
            app("B1 = %1.2f" % self.b1)
            fig.text(0.4, 0.5, "\n".join(text), transform=ax.transAxes)

        return fig



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This package is an overarching package for modules performing various analyses
of materials data.
"""




# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This module define a WulffShape class to generate the Wulff shape from
a lattice, a list of indices and their corresponding surface energies,
and the total area and volume of the wulff shape,the weighted surface energy,
the anisotropy and shape_factor can also be calculated.
In support of plotting from a given view in terms of miller index.

The lattice is from the conventional unit cell, and (hkil) for hexagonal lattices.
"""

from __future__ import division, unicode_literals
from pymatgen.core.structure import Structure
from pymatgen.core.surface import get_recp_symmetry_operation
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.coord_utils import get_angle
import numpy as np
import scipy as scp
from scipy.spatial import ConvexHull
import copy
import logging

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.colorbar as colorbar
import matplotlib.cm as cm
import mpl_toolkits.mplot3d as a3

__author__ = 'Zihan Xu, Richard Tran'
__copyright__ = 'Copyright 2013, The Materials Virtual Lab'
__version__ = '0.1'
__maintainer__ = 'Zihan Xu'
__email__ = 'zix009@eng.ucsd.edu'
__date__ = 'May 5 2016'

logger = logging.getLogger(__name__)


def hkl_tuple_to_str(hkl):
    """
    Prepare for display on plots
    "(hkl)" for surfaces
    Agrs:
        hkl: in the form of [h, k, l] or (h, k, l)
    """
    str_format = '($'
    for x in hkl:
        if x < 0:
            str_format += '\overline{' + str(-x) + '}'
        else:
            str_format += str(x)
    str_format += '$)'
    return str_format


def get_tri_area(pts):
    """
    Given a list of coords for 3 points,
    Compute the area of this triangle.
    Args:
        pts: [a, b, c] three points
    """
    a, b, c = pts[0], pts[1], pts[2]
    v1 = np.array(b) - np.array(a)
    v2 = np.array(c) - np.array(a)
    area_tri = abs(scp.linalg.norm(scp.cross(v1, v2)) / 2)
    return area_tri


class WulffShape(object):
    """
    Generate Wulff Shape from list of miller index and surface energies,
    with given conventional unit cell.
    surface energy (Jm^2) is the length of normal.

    Wulff shape is the convex hull.
    Based on:
    http://scipy.github.io/devdocs/generated/scipy.spatial.ConvexHull.html

    Process:
        1. get wulff simplices
        2. label with color
        3. get wulff_area and other properties

    .. attribute:: debug (bool)

    .. attribute:: alpha
        transparency

    .. attribute:: color_set

    .. attribute:: grid_off (bool)

    .. attribute:: axis_off (bool)

    .. attribute:: show_area

    .. attribute:: off_color
        color of planes off wulff

    .. attribute:: input_miller_fig
        ($input_miller$) for legend on figures

    .. attribute:: structure
        Structure object, input conventional unit cell (with H ) from lattice

    .. attribute:: input_miller
        list of input miller index, for hcp in the form of hkil

    .. attribute:: input_hkl
        modify hkill to hkl, in the same order with input_miller

    .. attribute:: input_e_surf
        list of input surface energies, in the same order with input_miller

    .. attribute:: latt
        Lattice object, the input lattice for the conventional unit cell

    .. attribute:: recp
        reciprocal_lattice_crystallographic of the input structure

    .. attribute:: recp_symmops
        list of symmetric operations for recp

    .. attribute:: cart_symmops
        list of symmetric operations for input structure (cartesian)

    .. attribute:: normal_e_m: item:
        [normal, e_surf, normal_pt, dual_pt, color_plane, m_ind_orig, miller]
        for all planes considering symm

    .. attribute:: dual_cv_simp
        simplices from the dual convex hull (dual_pt)

    .. attribute:: wulff_pt_list

    .. attribute:: wulff_cv_simp
        simplices from the convex hull of wulff_pt_list

    .. attribute:: simpx_info

    .. attribute:: plane_wulff_info: item:
        [normal, e_surf, pts, simpx, color_plane, m_ind_orig, miller]

    .. attribute:: on_wulff
        list for all input_miller, True is on wulff.

    .. attribute:: color_area
        list for all input_miller, total area on wulff, off_wulff = 0.

    .. attribute:: color_list
        color all input_miller, on_wulff ones according to e_surf

    .. attribute:: color_proxy
        color_proxy for all input_miller

    .. attribute:: color_proxy_on_wulff
        color_proxy for on wulff input_miller only

    .. attribute:: miller_on_wulff
        input_miller_fig for on wulff input_miller only

    .. attribute:: e_surf_on_wulff
        e_surf for on wulff input_miller only

    .. attribute:: miller_area
        ($hkl$): area for all input_miller

    """

    def __init__(self, lattice, miller_list, e_surf_list, color_set='PuBu',
                 grid_off=True, axis_off=True, show_area=False, alpha=1,
                 off_color='red', symprec=0.00001):
        """
        Args:
            lattice: Lattice object of the conventional unit cell
            miller_list: list of hkl or hkil for hcp
            e_surf_list: list of corresponding surface energies
            color_set: default is 'PuBu'
            grid_off(bool): default is True
            axis_off(bool): default is Ture
            show_area(bool): default is False
            alpha: chosen from 0 to 1 (float), default is 1
            off_color: color_legend for off_wulff planes on show_area legend
            symprec: for recp_operation, default is 0.01
        """
        latt = lattice.scale(1)
        structure = Structure(latt, ["H"], [[0, 0, 0]])
        # 1. store input args:
        # store plot settings:
        self.alpha = alpha
        self.color_set = color_set
        self.color_ind = list(range(len(miller_list)))
        self.grid_off = grid_off
        self.axis_off = axis_off
        self.show_area = show_area
        self.off_color = off_color
        self.input_miller_fig = [hkl_tuple_to_str(x) for x in miller_list]
        # store input data
        self.structure = structure
        self.input_miller = [list(x) for x in miller_list]
        self.input_hkl = copy.copy([[x[0], x[1], x[-1]] for x in miller_list])
        self.input_e_surf = copy.copy(e_surf_list)
        self.latt = latt
        self.recp = structure.lattice.reciprocal_lattice_crystallographic
        self.operation = get_recp_symmetry_operation
        self.recp_symmops = self.operation(structure, symprec)
        self.cart_symmops = self.symmop_cartesian(symprec)

        # 2. get all the data for wulff construction
        # get all the surface normal from get_all_miller_e()
        normal_e_m = self.get_all_miller_e()
        # [normal, e_surf, normal_pt, dual_pt, color_plane, m_ind_orig, miller]
        logger.debug(len(normal_e_m))
        self.normal_e_m = normal_e_m

        # 3. consider the dual condition
        dual_pts = [x[3] for x in normal_e_m]
        dual_convex = ConvexHull(dual_pts)
        dual_cv_simp = dual_convex.simplices
        # simplices	(ndarray of ints, shape (nfacet, ndim))
        # list of [i, j, k] , ndim = 3
        # i, j, k: ind for normal_e_m
        # recalculate the dual of dual, get the wulff shape.
        # conner <-> surface
        # get cross point from the simplices of the dual convex hull
        wulff_pt_list = [self.get_cross_pt_dual_simp(dual_simp)
                         for dual_simp in dual_cv_simp]

        wulff_convex = ConvexHull(wulff_pt_list)
        wulff_cv_simp = wulff_convex.simplices
        logger.debug(", ".join([str(len(x)) for x in wulff_cv_simp]))

        # store simplices and convex
        self.dual_cv_simp = dual_cv_simp
        self.wulff_pt_list = wulff_pt_list
        self.wulff_cv_simp = wulff_cv_simp

        # 4. get wulff info
        # return (simpx_info, plane_wulff_info, on_wulff, surface_area)
        wulff_info = self.get_simpx_plane()
        self.simpx_info = wulff_info[0]
        # need to update the color
        # plan_wulff_info: [normal, e_surf, pts, simpx,
        #   color_plane, m_ind_orig, miller]
        self.plane_wulff_info = wulff_info[1]
        self.on_wulff = wulff_info[2]
        self.color_area = wulff_info[3]

        # 5. assign color for on_wulff plane
        # return (color_list, color_proxy, color_proxy_on_wulff,
        # miller_on_wulff, e_surf_on_wulff_list)
        color_info = self.get_colors()
        self.color_list = color_info[0]
        self.color_proxy = color_info[1]
        self.color_proxy_on_wulff = color_info[2]
        self.miller_on_wulff = color_info[3]
        self.e_surf_on_wulff = color_info[4]

        miller_area = []
        for m, in_mill_fig in enumerate(self.input_miller_fig):
            miller_area.append(
                in_mill_fig + ' : ' + str(round(self.color_area[m], 4)))
        self.miller_area = miller_area

    def symmop_cartesian(self, symmprec):
        structure = self.structure
        space_group_analyzer = SpacegroupAnalyzer(structure, symmprec)
        symm_ops = space_group_analyzer.get_point_group_operations(
            cartesian=True)
        return symm_ops

    def get_all_miller_e(self):
        """
        from self:
            get miller_list(unique_miller), e_surf_list and symmetry
            operations(symmops) according to lattice
        apply symmops to get all the miller index, then get normal,
        get all the planes functions for wulff shape calculation:
            |normal| = 1, e_surf is plane's distance to (0, 0, 0),
            normal[0]x + normal[1]y + normal[2]z = e_surf

        return:
            normal_e_m, item: [normal, e_surf, normal_pt, dual_pt,
            color_plane, m_ind_orig, miller]
        """
        all_hkl = copy.copy(self.input_hkl)
        all_hkl_ind = list(enumerate(all_hkl))
        e_surf_list = copy.copy(self.input_e_surf)
        symmops = self.recp_symmops
        recp = self.recp
        color_ind = self.color_ind
        normal_e_m = []
        color = copy.copy(color_ind)
        miller_ind_orig = [x[0] for x in all_hkl_ind]

        for i, hkl in enumerate(all_hkl):
            for op in symmops:
                miller = list(op.operate(hkl))
                miller = [int(x) for x in miller]
                if miller in all_hkl:
                    continue
                else:
                    all_hkl.append(miller)
                    e_surf_list.append(e_surf_list[i])
                    miller_ind_orig.append(i)
                    color.append(color_ind[divmod(i, len(color_ind))[1]])

        for i, hkl in enumerate(all_hkl):
            # get normal (length=1)
            normal = recp.get_cartesian_coords(hkl)
            normal /= scp.linalg.norm(normal)
            e_surf = e_surf_list[i]
            normal_pt = [x * e_surf for x in normal]
            dual_pt = [x / e_surf for x in normal]
            # the index for color and plane
            color_plane = color[i]
            m_ind_orig = miller_ind_orig[i]
            normal_e_m.append([normal, e_surf, normal_pt, dual_pt,
                               color_plane, m_ind_orig, hkl])

        # sorted by e_surf
        normal_e_m.sort(key=lambda x: x[1])
        return normal_e_m

    def get_cross_pt_dual_simp(self, dual_simp):
        """
        |normal| = 1, e_surf is plane's distance to (0, 0, 0),
        plane function:
            normal[0]x + normal[1]y + normal[2]z = e_surf
        normal_e_m, item: [normal, e_surf, normal_pt, dual_pt,
            color_plane, m_ind_orig, miller]

        from self:
            normal_e_m to get the plane functions
            dual_simp: (i, j, k) simplices from the dual convex hull
                i, j, k: plane index(same order in normal_e_m)
        """
        i, j, k = dual_simp[0], dual_simp[1], dual_simp[2]
        normal_e_m = self.normal_e_m
        matrix_surfs = [normal_e_m[i][0], normal_e_m[j][0], normal_e_m[k][0]]
        matrix_e = [normal_e_m[i][1], normal_e_m[j][1], normal_e_m[k][1]]
        cross_pt = scp.dot(scp.linalg.inv(matrix_surfs), matrix_e)
        return cross_pt

    def get_simpx_plane(self):
        """
        local the plane for simpx of on wulff_cv,
        by comparing the center of the simpx triangle
        with the plane functions.

        based on: wulff_cv_simp, normal_e_m, wulff_pt_list
        """
        wulff_simpx = self.wulff_cv_simp
        # normal_e_m, item: [normal, e_surf, normal_pt, dual_pt,
        #     color_plane, m_ind_orig, miller]
        normal_e_m = self.normal_e_m
        wulff_pt_list = self.wulff_pt_list
        simpx_info = []
        on_wulff = [False] * len(self.input_miller)
        surface_area = [0.0] * len(self.input_miller)
        plane_wulff_info = [[x[0], x[1], [], [], x[4], x[5], x[6]] for x in
                            normal_e_m]
        # each simpx (i,j,k) from self.wulff_cv_simp
        #  forms a triangle on the wulff shape.
        # check which surface it belongs to
        for simpx in wulff_simpx:
            pts = [wulff_pt_list[simpx[0]], wulff_pt_list[simpx[1]],
                   wulff_pt_list[simpx[2]]]
            center = np.sum(pts, 0) / 3.0
            # check whether the center of the simplices is on one plane
            for i, plane in enumerate(normal_e_m):
                normal, e_surf = plane[0], plane[1]
                abs_diff = abs(np.dot(normal, center) - e_surf)
                if abs_diff < 10e-6:
                    # assign values only if the simplices is on the plane.
                    plane_ind = plane[4]  # the ind of input_miller
                    on_wulff[plane_ind] = True
                    surface_area[plane_ind] += get_tri_area(pts)
                    simp_inf = [plane[0], plane[1], pts, simpx, plane[4],
                                plane[5], plane[6]]
                    # build a list in the same order of simpx
                    # with related plane information
                    simpx_info.append(simp_inf)

                    plane_wulff_info[i][2].append(pts)
                    # outer_lines
                    plane_wulff_info[i][3].append([simpx[0], simpx[1]])
                    plane_wulff_info[i][3].append([simpx[1], simpx[2]])
                    plane_wulff_info[i][3].append([simpx[0], simpx[2]])
                    # already find the plane, move to the next simplices
                    break
        for plane in plane_wulff_info:
            if len(plane[3]):
                plane[3].sort()
                outer_lines = []
                for line in plane[3]:
                    if plane[3].count(line) == 2:
                        continue
                    outer_lines.append(line)
                plane[3] = outer_lines
        # plan_wulff_info: [normal, e_surf, pts, simpx,
        #     color_plane, m_ind_orig, miller]
        return simpx_info, plane_wulff_info, on_wulff, surface_area

    def get_colors(self):
        """
        assign colors according to the surface energies of on_wulff planes.

        return:
            (color_list, color_proxy, color_proxy_on_wulff, miller_on_wulff,
            e_surf_on_wulff_list)
        """
        on_wulff = self.on_wulff
        input_hkl = self.input_hkl
        input_e_surf = copy.copy(self.input_e_surf)
        input_miller_fig = self.input_miller_fig
        color_set = self.color_set
        alpha = self.alpha
        color_list = [self.off_color] * len(input_hkl)
        color_proxy_on_wulff = []
        miller_on_wulff = []
        e_surf_on_wulff = []
        # get list of on_wulff surface energies (with ind)
        for i, e_surf in enumerate(input_e_surf):
            if on_wulff[i]:
                e_surf_on_wulff.append((i, e_surf))

        c_map = plt.get_cmap(color_set)
        e_surf_on_wulff.sort(key=lambda x: x[1], reverse=False)
        e_surf_on_wulff_list = [x[1] for x in e_surf_on_wulff]
        if len(e_surf_on_wulff) > 1:
            cnorm = colors.Normalize(vmin=min(e_surf_on_wulff_list),
                                     vmax=max(e_surf_on_wulff_list))
        else:
            # if there is only one hkl on wulff, choose the color of the median
            cnorm = colors.Normalize(vmin=min(e_surf_on_wulff_list) - 0.1,
                                     vmax=max(e_surf_on_wulff_list) + 0.1)
        scalar_map = cm.ScalarMappable(norm=cnorm, cmap=c_map)
        logger.debug(e_surf_on_wulff)
        # prepare for
        for i, e_surf in e_surf_on_wulff:
            plane_color = scalar_map.to_rgba(e_surf, alpha=alpha)
            color_list[i] = plane_color
            color_proxy_on_wulff.append(
                plt.Rectangle((2, 2), 1, 1, fc=plane_color, alpha=alpha))
            miller_on_wulff.append(input_miller_fig[i])
        scalar_map.set_array([x[1] for x in e_surf_on_wulff])
        color_proxy = [plt.Rectangle((2, 2), 1, 1, fc=x, alpha=alpha)
                       for x in color_list]

        return color_list, color_proxy, color_proxy_on_wulff, miller_on_wulff, \
            e_surf_on_wulff_list

    def plot_wf_simpx(self, direction=None, bar_pos=(0.75, 0.15, 0.05, 0.65),
                      bar_on=False, legend_on=True, aspect_ratio=(8, 8)):
        """
        plot the wulff shape from self.wulff_pt_list, self.plane_wulff_info

        Args:
            direction: default is (1, 1, 1)
            bar_pos: default is [0.75, 0.15, 0.05, 0.65]
            bar_on (bool): default is False
            legend_on (bool): default is True
            aspect_ratio: default is (8, 8)

        """

        if not direction:
            area_dict = self.area_fraction_dict
            miller_indices, areas = [], []
            for hkl in area_dict.keys():
                miller_indices.append(hkl)
                areas.append(area_dict[hkl])
            direction = miller_indices[areas.index(max(areas))]
        fig = plt.figure()
        fig.set_size_inches(aspect_ratio[0],
                            aspect_ratio[1])
        azim, elev = self.get_azimuth_elev(direction)

        wulff_pt_list = self.wulff_pt_list
        plane_wulff_info = self.plane_wulff_info
        # [normal, e_surf, [pts], [simpx],
        #     color_plane, m_ind_orig, miller]
        ax = a3.Axes3D(fig, azim=azim, elev=elev)

        for plane in plane_wulff_info:
            # check whether [pts] is empty
            if len(plane[2]) < 1:
                # empty, plane is not on_wulff.
                continue
            # assign the color for on_wulff planes according to its
            # color_plane index and the color_list for on_wulff
            plane_color = self.color_list[plane[4]]
            # plane[3]: [simpx]
            lines = list(plane[3])
            pt = []
            prev = None
            while len(lines) > 0:
                if prev is None:
                    l = lines.pop(0)
                else:
                    for i, l in enumerate(lines):
                        if prev in l:
                            l = lines.pop(i)
                            if l[1] == prev:
                                l.reverse()
                            break
                # make sure the lines are connected one by one.
                # find the way covering all pts and facets
                pt.append(self.wulff_pt_list[l[0]].tolist())
                pt.append(self.wulff_pt_list[l[1]].tolist())
                prev = l[1]
            # plot from the sorted pts from [simpx]
            tri = a3.art3d.Poly3DCollection([pt])
            tri.set_color(plane_color)
            # "#808080" is the default edge color.
            tri.set_edgecolor("#808080")
            ax.add_collection3d(tri)
        # set ranges of x, y, z
        # find the largest distance between on_wulff pts and the origin,
        # to ensure complete and consistent display for all directions
        r_range = max([np.linalg.norm(x) for x in wulff_pt_list])
        ax.set_xlim([-r_range * 1.1, r_range * 1.1])
        ax.set_ylim([-r_range * 1.1, r_range * 1.1])
        ax.set_zlim([-r_range * 1.1, r_range * 1.1])
        # add legend
        if legend_on:
            color_proxy = self.color_proxy
            if self.show_area:
                ax.legend(color_proxy, self.miller_area, loc='upper left',
                          bbox_to_anchor=(0, 1), fancybox=True, shadow=False)
            else:
                ax.legend(self.color_proxy_on_wulff, self.miller_on_wulff,
                          loc='upper center',
                          bbox_to_anchor=(0.5, 1), ncol=3, fancybox=True,
                          shadow=False)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        # add colorbar
        # cbar = fig.colorbar(self.scalarcm, alpha=self.alpha)
        # Add an axes at position rect [left, bottom, width, height]
        cmap = plt.get_cmap(self.color_set)
        cmap.set_over('0.25')
        cmap.set_under('0.75')
        bounds = [round(e, 2) for e in self.e_surf_on_wulff]
        bounds.append(1.2 * bounds[-1])
        norm = colors.BoundaryNorm(bounds, cmap.N)
        if bar_on:
            # display surface energies
            ax1 = fig.add_axes(bar_pos)
            cbar = colorbar.ColorbarBase(ax1, cmap=cmap, norm=norm,
                                         boundaries=[0] + bounds + [10],
                                         extend='both',
                                         ticks=bounds[:-1],  # optional
                                         spacing='proportional',
                                         orientation='vertical'
                                         )
            cbar.set_label('Surface Energies ($J/m^2$)', fontsize=100)
        # [normal, e_surf, normal_pt, dual_pt, color_plane, m_ind_orig, miller]
        if self.grid_off:
            ax.grid('off')
        if self.axis_off:
            ax.axis('off')
        plt.draw()
        return plt

    def get_azimuth_elev(self, miller_index):
        """
        :param miller_index: viewing direction
        :return: azim, elev for plotting
        """

        cart = self.latt.get_cartesian_coords(miller_index)
        azim = get_angle([cart[0], cart[1], 0], (1, 0, 0))
        v = [cart[0], cart[1], 0]
        elev = get_angle(cart, v)
        if miller_index == (0, 0, 1) or miller_index == (0, 0, 0, 1):
            return 0, 90
        else:
            return azim, elev

    @property
    def wulff_volume(self):
        """
        :return:
            the volume of the wulff shape
        """
        return ConvexHull(self.wulff_pt_list).volume

    @property
    def miller_area_dict(self):
        """
        :return:
            (dict): {hkl: area_hkl on wulff}
        """
        miller_area_dict = {}
        for i, hkl in enumerate(self.input_miller):
            miller_area_dict[tuple(hkl)] = self.color_area[i]
        return miller_area_dict

    @property
    def miller_energy_dict(self):
        """
        :return:
            (dict): {hkl: surface energy_hkl}
        """
        miller_energy_dict = {}
        for i, hkl in enumerate(self.input_miller):
            miller_energy_dict[tuple(hkl)] = self.input_e_surf[i]
        return miller_energy_dict

    @property
    def total_surface_area(self):
        """
        :return:
            total area on wulff
        """
        tot_area = 0
        for hkl in self.miller_area_dict.keys():
            tot_area += self.miller_area_dict[hkl]
        return tot_area

    @property
    def weighted_surface_energy(self):
        """
        :return:
            sum(surface_energy_hkl * area_hkl)/ sum(area_hkl)
        """
        tot_area_energy = 0
        for hkl in self.miller_energy_dict.keys():
            tot_area_energy += self.miller_energy_dict[hkl] * \
                               self.miller_area_dict[hkl]
        return tot_area_energy / self.total_surface_area

    @property
    def area_fraction_dict(self):
        """
        :return:
            (dict): {hkl: area_hkl/total area on wulff}
        """
        area_fraction_dict = {}
        for hkl in self.miller_area_dict.keys():
            area_fraction_dict[hkl] = self.miller_area_dict[hkl] / \
                                      self.total_surface_area
        return area_fraction_dict

    @property
    def anisotropy(self):
        """
        :return:
            variation from weighted surface energy
            The ideal sphere is 0.
        """
        square_diff_energy = 0
        weighted_energy = self.weighted_surface_energy
        area_frac_dict = self.area_fraction_dict
        miller_energy_dict = self.miller_energy_dict

        for hkl in miller_energy_dict.keys():
            square_diff_energy += ((miller_energy_dict[
                                        hkl] - weighted_energy) ** 2) * \
                                  area_frac_dict[hkl]
        return np.sqrt(square_diff_energy) / weighted_energy

    @property
    def shape_factor(self):
        """
        This is useful for determining the critical nucleus size.
        A large shape factor indicates great anisotropy.
        See Ballufi, R. W., Allen, S. M. & Carter, W. C. Kinetics
            of Materials. (John Wiley & Sons, 2005), p.461

        :return:
            variation from weighted surface energy
            The ideal sphere is 0.
        """
        return self.total_surface_area / (self.wulff_volume ** (2 / 3))



from __future__ import unicode_literals

__author__ = "Pymatgen Development Team"
__email__ ="pymatgen@googlegroups.com"
__maintainer__ = "Shyue Ping Ong"
__maintainer_email__ ="shyuep@gmail.com"
__date__ = "Jul 14 2016"
__version__ = "4.1.0"


# Order of imports is important on some systems to avoid 
# failures when loading shared libraries.
import spglib
from . import optimization, util
del(spglib, optimization, util)

# Useful aliases for commonly used objects and modules.
# Allows from pymatgen import <class> for quick usage.

from .core import *
from .electronic_structure.core import Spin, Orbital
from .matproj.rest import MPRester
from monty.json import MontyEncoder, MontyDecoder, MSONable



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
The defects package implements defect generation tools
"""

from .point_defects import *



# coding: utf-8
"""Mock objects for unit tests."""
from __future__ import division, print_function, unicode_literals

from .nodes import Status
from .tasks import AbinitTask
from .flows import Flow


def change_task_start(task, mocked_status="Error"):
    """Return a AbinitTaskMockedStart object."""
    task.__class__ = AbinitTaskMockedStart
    task.mocked_status = Status.as_status(mocked_status)
    return task


class AbinitTaskMockedStart(AbinitTask):
    """A Task whose status is always self.mocked_status."""
    def start(self, **kwargs):
        self.set_status(self.mocked_status, msg="Mocking status with %s" % self.mocked_status)
        return 1


def infinite_flow(flow):
    """Return an InfiniteFlow."""
    flow.__class__ = InfiniteFlow
    return flow


class InfiniteFlow(Flow):
    """A Flow that will never reach `all_ok`"""
    def check_status(self, **kwargs):
        super(InfiniteFlow, self).check_status(**kwargs)

        for task in self.iflat_tasks(status=self.S_OK):
            task.set_status(task.S_INIT)
            task.reset_from_scratch()
